from .beta_distribution import get_beta_probabilities
from .binomial_distribution import get_binomial_probabilities
from .exponential_distribution import get_exponential_probabilities
from .general_triangular_distribution import get_general_triangular_probabilities
from .poisson_distribution import get_poisson_probabilities
from .triangular_distribution import get_triangular_probabilities

__all__ = [
    "get_beta_probabilities",
    "get_exponential_probabilities",
    "get_binomial_probabilities",
    "get_poisson_probabilities",
    "get_triangular_probabilities",
    "get_general_triangular_probabilities",
]



from typing import Optional

import torch
from torch import Tensor

from ..distributions import get_beta_probabilities
from .custom_targets_loss import CustomTargetsCrossEntropyLoss

# Params [a,b] for beta distribution
_beta_params_sets = {
    "standard": {
        3: [[1, 4, 1], [4, 4, 1], [4, 1, 1]],
        4: [[1, 6, 1], [6, 10, 1], [10, 6, 1], [6, 1, 1]],
        5: [[1, 8, 1], [6, 14, 1], [12, 12, 1], [14, 6, 1], [8, 1, 1]],
        6: [[1, 10, 1], [7, 20, 1], [15, 20, 1], [20, 15, 1], [20, 7, 1], [10, 1, 1]],
        7: [
            [1, 12, 1],
            [7, 26, 1],
            [16, 28, 1],
            [24, 24, 1],
            [28, 16, 1],
            [26, 7, 1],
            [12, 1, 1],
        ],
        8: [
            [1, 14, 1],
            [7, 31, 1],
            [17, 37, 1],
            [27, 35, 1],
            [35, 27, 1],
            [37, 17, 1],
            [31, 7, 1],
            [14, 1, 1],
        ],
        9: [
            [1, 16, 1],
            [8, 40, 1],
            [18, 47, 1],
            [30, 47, 1],
            [40, 40, 1],
            [47, 30, 1],
            [47, 18, 1],
            [40, 8, 1],
            [16, 1, 1],
        ],
        10: [
            [1, 18, 1],
            [8, 45, 1],
            [19, 57, 1],
            [32, 59, 1],
            [45, 55, 1],
            [55, 45, 1],
            [59, 32, 1],
            [57, 19, 1],
            [45, 8, 1],
            [18, 1, 1],
        ],
        11: [
            [1, 21, 1],
            [8, 51, 1],
            [20, 68, 1],
            [34, 73, 1],
            [48, 69, 1],
            [60, 60, 1],
            [69, 48, 1],
            [73, 34, 1],
            [68, 20, 1],
            [51, 8, 1],
            [21, 1, 1],
        ],
        12: [
            [1, 23, 1],
            [8, 56, 1],
            [20, 76, 1],
            [35, 85, 1],
            [51, 85, 1],
            [65, 77, 1],
            [77, 65, 1],
            [85, 51, 1],
            [85, 35, 1],
            [76, 20, 1],
            [56, 8, 1],
            [23, 1, 1],
        ],
        13: [
            [1, 25, 1],
            [8, 61, 1],
            [20, 84, 1],
            [36, 98, 1],
            [53, 100, 1],
            [70, 95, 1],
            [84, 84, 1],
            [95, 70, 1],
            [100, 53, 1],
            [98, 36, 1],
            [84, 20, 1],
            [61, 8, 1],
            [25, 1, 1],
        ],
        14: [
            [1, 27, 1],
            [2, 17, 1],
            [5, 23, 1],
            [9, 27, 1],
            [13, 28, 1],
            [18, 28, 1],
            [23, 27, 1],
            [27, 23, 1],
            [28, 18, 1],
            [28, 13, 1],
            [27, 9, 1],
            [23, 5, 1],
            [17, 2, 1],
            [27, 1, 1],
        ],
    }
}


class BetaCrossEntropyLoss(CustomTargetsCrossEntropyLoss):
    """Beta unimodal regularised cross entropy loss from :footcite:t:`vargas2022unimodal`.

    Parameters
    ----------
    num_classes : int, default=5
        Number of classes.
    params_set : str, default='standard'
        The set of parameters to use for the beta distribution (chosen from the
        _beta_params_set dictionary).
    eta : float, default=1.0
        Parameter that controls the influence of the regularisation.
    weight : Optional[Tensor], default=None
        A manual rescaling weight given to each class. If given, has to be a Tensor
        of size `C`. Otherwise, it is treated as if having all ones.
    size_average : Optional[bool], default=None
        Deprecated (see :attr:`reduction`). By default, the losses are averaged over
        each loss element in the batch. Note that for some losses, there are
        multiple elements per sample. If the field :attr:`size_average` is set to
        ``False``, the losses are instead summed for each minibatch. Ignored when
        reduce is ``False``. Default: ``True``
    ignore_index : int, default=-100
        Specifies a target value that is ignored and does not contribute to the
        input gradient. When :attr:`size_average` is ``True``, the loss is averaged
        over non-ignored targets.
    reduce : Optional[bool], default=None
        Deprecated (see :attr:`reduction`). By default, the losses are averaged or
        summed over observations for each minibatch depending on :attr:`size_average`.
        When :attr:`reduce` is ``False``, returns a loss per batch element instead
        and ignores :attr:`size_average`. Default: ``True``
    reduction : str, default='mean'
        Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` |
        ``'sum'``. ``'none'``: no reduction will be applied, ``'mean'``: the sum of
        the output will be divided by the number of elements in the output,
        ``'sum'``: the output will be summed. Note: :attr:`size_average` and
        :attr:`reduce` are in the process of being deprecated, and in the meantime,
        specifying either of those two args will override :attr:`reduction`.
        Default: ``'mean'``
    label_smoothing : float, default=0.0
        Controls the amount of label smoothing for the loss. Zero means no smoothing.
        Default: ``0.0``
    """

    def __init__(
        self,
        num_classes: int = 5,
        params_set: str = "standard",
        eta: float = 1.0,
        weight: Optional[Tensor] = None,
        size_average=None,
        ignore_index: int = -100,
        reduce=None,
        reduction: str = "mean",
        label_smoothing: float = 0.0,
    ):
        self.params = _beta_params_sets[params_set]

        # Precompute class probabilities for each label
        cls_probs = torch.tensor(
            [
                get_beta_probabilities(
                    num_classes,
                    self.params[num_classes][i][0],
                    self.params[num_classes][i][1],
                    self.params[num_classes][i][2],
                )
                for i in range(num_classes)
            ]
        ).float()

        super().__init__(
            cls_probs=cls_probs,
            eta=eta,
            weight=weight,
            size_average=size_average,
            ignore_index=ignore_index,
            reduce=reduce,
            reduction=reduction,
            label_smoothing=label_smoothing,
        )



from typing import Dict, Callable
from torch import nn

activation_function_by_name: Dict[str, Callable[[], nn.Module]] = {
    "relu": lambda: nn.ReLU(inplace=True),
    "elu": lambda: nn.ELU(inplace=True),
    "softplus": lambda: nn.Softplus(),
}



import numpy as np
from scipy.stats import binom


def get_binomial_probabilities(n):
    """Get probabilities from binominal distribution for n classes.

    Parameters
    ----------
    n : int
            Number of classes.

    Returns
    -------
    probs : 2d array-like
            Matrix of probabilities where each row represents the true class
            and each column the probability for class n.

    Example
    -------
    >>> get_binominal_probabilities(5)
    array([[6.561e-01, 2.916e-01, 4.860e-02, 3.600e-03, 1.000e-04],
            [2.401e-01, 4.116e-01, 2.646e-01, 7.560e-02, 8.100e-03],
            [6.250e-02, 2.500e-01, 3.750e-01, 2.500e-01, 6.250e-02],
            [8.100e-03, 7.560e-02, 2.646e-01, 4.116e-01, 2.401e-01],
            [1.000e-04, 3.600e-03, 4.860e-02, 2.916e-01, 6.561e-01]])
    """

    params = {}

    params["4"] = np.linspace(0.1, 0.9, 4)
    params["5"] = np.linspace(0.1, 0.9, 5)
    params["6"] = np.linspace(0.1, 0.9, 6)
    params["7"] = np.linspace(0.1, 0.9, 7)
    params["8"] = np.linspace(0.1, 0.9, 8)
    params["10"] = np.linspace(0.1, 0.9, 10)
    params["12"] = np.linspace(0.1, 0.9, 12)
    params["14"] = np.linspace(0.1, 0.9, 14)

    probs = []

    for true_class in range(0, n):
        probs.append(binom.pmf(np.arange(0, n), n - 1, params[str(n)][true_class]))

    return np.array(probs)



from math import sqrt

import torch
from torch.nn import Module


class CLM(Module):
    """
    Implementation of the cumulative link model from :footcite:t:`vargas2020clm` as a torch layer.
    Different link functions can be used, including logit, probit and cloglog.

    Parameters
    ----------
    num_classes : int
        The number of classes.
    link_function : str
        The link function to use. Can be ``'logit'``, ``'probit'`` or ``'cloglog'``.
    min_distance : float, default=0.0
       The minimum distance between thresholds
    """

    def __init__(self, num_classes, link_function, min_distance=0.0, **kwargs):
        super().__init__()
        self.num_classes = num_classes
        self.link_function = link_function
        self.min_distance = min_distance
        self.dist = torch.distributions.Normal(0, 1)
        self.device = "cpu"

        self.thresholds_b = torch.nn.Parameter(data=torch.Tensor(1), requires_grad=True)
        torch.nn.init.uniform_(self.thresholds_b, 0.0, 0.1)

        self.thresholds_a = torch.nn.Parameter(
            data=torch.Tensor(self.num_classes - 2), requires_grad=True
        )
        torch.nn.init.uniform_(
            self.thresholds_a,
            sqrt((1.0 / (self.num_classes - 2)) / 2),
            sqrt(1.0 / (self.num_classes - 2)),
        )

    def _convert_thresholds(self, b, a, min_distance):
        a = a**2
        a = a + min_distance
        thresholds_param = torch.cat((b, a), dim=0)
        th = torch.sum(
            torch.tril(
                torch.ones((self.num_classes - 1, self.num_classes - 1)).to(
                    self.device
                ),
                diagonal=-1,
            )
            * torch.reshape(
                torch.tile(thresholds_param, (self.num_classes - 1,)).to(self.device),
                shape=(self.num_classes - 1, self.num_classes - 1),
            ),
            dim=(1,),
        )
        return th

    def _clm(self, projected: torch.Tensor, thresholds: torch.Tensor):
        projected = torch.reshape(projected, shape=(-1,))

        m = projected.shape[0]
        a = torch.reshape(torch.tile(thresholds, (m,)), shape=(m, -1))
        b = torch.transpose(
            torch.reshape(
                torch.tile(projected, (self.num_classes - 1,)), shape=(-1, m)
            ),
            0,
            1,
        )
        z3 = a - b

        if self.link_function == "probit":
            a3T = self.dist.cdf(z3)
        elif self.link_function == "cloglog":
            a3T = 1 - torch.exp(-torch.exp(z3))
        else:
            a3T = 1.0 / (1.0 + torch.exp(-z3))

        ones = torch.ones((m, 1)).to(self.device)
        a3 = torch.cat((a3T, ones), dim=1)
        a3[:, 1:] = a3[:, 1:] - a3[:, 0:-1]

        return a3

    def forward(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor
            The input tensor.

        Returns
        -------
        output: Tensor
            The output tensor.
        """

        thresholds = self._convert_thresholds(
            self.thresholds_b, self.thresholds_a, self.min_distance
        )

        return self._clm(x, thresholds)

    def to(self, device):
        self.device = device

        return self



from scipy.special import gamma, hyp2f1

from .utils import get_intervals


def beta_func(a, b):
    """Compute the beta function.

    Parameters
    ----------
    a : float
            First parameter.
    b : float
            Second parameter.

    Returns
    -------
    beta_func : float
            The value of the beta function.
    """

    if a <= 0:
        raise ValueError(f"{a=} can not be negative")

    if b <= 0:
        raise ValueError(f"{b=} can not be negative")

    return (gamma(a) * gamma(b)) / gamma(a + b)


def beta_dist(x, p, q, a=1.0):
    """Compute the beta distribution function for value x with parameters p, q and a.

    Parameters
    ----------
    x : float
            Value to compute the distribution function for.
    p : float
            First shape parameter (:math:`p > 0`).
    q : float
            Second shape parameter (:math:`q > 0`).
    a : float, default=1.0
            Scaling parameter.

    Returns
    -------
    beta: float
            The value of the beta distribution function.
    """

    if x < 0 or x > 1:
        raise ValueError(f"{x=} must be in the interval [0,1]")

    if a <= 0:
        raise ValueError(f"{a=} can not be negative or 0")

    return (x ** (a * p)) / (p * beta_func(p, q)) * hyp2f1(p, 1 - q, p + 1, x**a)


def get_beta_probabilities(n, p, q, a=1.0):
    """Get probabilities from a beta distribution :math:`B(p,q,a)` for ``n`` splits.

    Parameters
    ----------
    n : int
            Number of classes.
    p : float
            First shape parameter (:math:`p > 0`).
    q : float
            Second shape parameter (:math:`q > 0`).
    a : float, default=1.0
            Scaling parameter.

    Returns
    -------
    probs: list
            List of probabilities.
    """

    intervals = get_intervals(n)
    probs = []

    # Compute probability for each interval (class) using the distribution function.
    for interval in intervals:
        probs.append(beta_dist(interval[1], p, q, a) - beta_dist(interval[0], p, q, a))

    return probs



import math

import numpy as np
from sympy import Eq, nsolve, solve, symbols
from sympy.core.add import Add

from .utils import get_intervals, triangular_cdf


def get_general_triangular_params(n: int, alphas: np.ndarray, verbose: int = 0):
    """
    Get the parameters for the general triangular distribution.

    Parameters
    ----------
    n : int
        Number of classes.
    alphas : np.ndarray
        Array of alphas.
    verbose : int, optional
        Verbosity level, by default 0.
    """
    alphas = np.array(alphas)
    if not isinstance(alphas, np.ndarray) or alphas.shape != (2 * n,):
        raise ValueError(
            f"alphas must be a numpy array of shape (2 * n,), but got {alphas.shape}"
        )

    def abc1(n, alpha):
        a = 0
        b = 1.0 / ((1.0 - math.sqrt(alpha)) * n)
        c = 0
        return a, b, c

    def abcJ(n, alpha):
        a = 1.0 - 1.0 / (n * (1.0 - math.sqrt(alpha)))
        b = 1
        c = 1
        return a, b, c

    def abcj(n, j, alpha2j_1, alpha2j):
        # Calculation of the terms of the system of equations
        # b = ...
        numa2 = 2 * n**2 - 2 * n**2 * alpha2j_1
        numa1 = 2 * j * n * alpha2j_1 - n * alpha2j_1 - 4 * n * j + 4 * n
        numa0 = 2 * j**2 - 4 * j + 2
        dena0 = 2 * j * n * alpha2j_1 - n * alpha2j_1
        dena1 = -(2 * n**2 * alpha2j_1)

        # a = ...
        numb2 = 2 * n**2 * alpha2j - 2 * n**2
        numb1 = 4 * n * j - 2 * n * j * alpha2j + n * alpha2j
        numb0 = -(2 * j**2)
        denb1 = 2 * n**2 * alpha2j
        denb0 = -2 * n * j * alpha2j + n * alpha2j

        a, b = symbols("a, b")
        c = (2 * j - 1) / (2 * n)
        eq1 = Eq((numa2 * a**2 + numa1 * a + numa0) / (dena0 + dena1 * a), b)
        eq2 = Eq((numb2 * b**2 + numb1 * b + numb0) / (denb1 * b + denb0), a)

        try:
            nsol = nsolve([eq1, eq2], [a, b], [1, 1])

            if nsol[0] < (j - 1) / n and nsol[1] > j / n:
                return nsol[0], nsol[1], c
        except ValueError:
            pass

        try:
            sol = solve([eq1, eq2], [a, b])
        except ValueError:
            raise ValueError(
                f"Unsatisfiable alpha values {alphas}: could not solve for the"
                " triangular distribution parameters"
            )

        soln = [tuple(v.evalf() for v in s) for s in sol]

        for s in soln:
            s_a = s[0]
            s_b = s[1]

            if isinstance(s_a, Add):
                s_a = s_a.args[0]
            if isinstance(s_b, Add):
                s_b = s_b.args[0]

            if s_a < (j - 1) / n and s_b > j / n:
                return s_a, s_b, c

        raise ValueError(f"Could not find solution for {j=}, {alpha2j_1=}, {alpha2j=}")

    if verbose >= 3:
        print(f"{abc1(n,alphas[2])=}, {abcJ(n,alphas[2*n-1])=}")
        for i in range(2, n):
            print(f"{i=}  {abcj(n,i,alphas[2*i-1],alphas[2*i])}")

    params = []

    # Compute params for each interval (class)
    for j in range(1, n + 1):
        j_params = {}
        if j == 1:
            a, b, c = abc1(n, alphas[2])
            j_params["alpha2j_1"] = 0
            j_params["alpha2j"] = alphas[2]
        elif j == n:
            a, b, c = abcJ(n, alphas[2 * n - 1])
            j_params["alpha2j_1"] = alphas[2 * n - 1]
            j_params["alpha2j"] = 0
        else:
            a, b, c = abcj(n, j, alphas[2 * j - 1], alphas[2 * j])
            j_params["alpha2j_1"] = alphas[2 * j - 1]
            j_params["alpha2j"] = alphas[2 * j]

        j_params["a"] = a
        j_params["b"] = b
        j_params["c"] = c
        params.append(j_params)

    return params


def get_general_triangular_probabilities(n: int, alphas: np.ndarray, verbose: int = 0):
    """
    Get the probabilities for the general triangular distribution.

    Parameters
    ----------
    n : int
        Number of classes.
    alphas : np.ndarray
        Array of alphas.
    verbose : int, optional
        Verbosity level, by default 0.
    """

    intervals = get_intervals(n)
    probs = []

    # Compute probability for each interval (class) using the distribution function.
    params = get_general_triangular_params(n, alphas, verbose)

    for param in params:
        a = param["a"]
        b = param["b"]
        c = param["c"]
        j_probs = []

        for interval in intervals:
            j_probs.append(
                triangular_cdf(interval[1], a, b, c)
                - triangular_cdf(interval[0], a, b, c)
            )
            if verbose >= 2:
                print(f"\tinterval: {interval}, prob={j_probs[-1]}")

        probs.append(j_probs)

    return np.array(probs, dtype=np.float32)



from typing import Optional, Union

import torch
import torch.nn.functional as F
from sklearn.base import BaseEstimator
from torch.utils.data import DataLoader


class PytorchEstimator(BaseEstimator):
    """
    Wrapper around a Pytorch ``nn.Module`` implementing
    the default estimator interface defined by ``scikit-learn``.

    Parameters
    ----------
    model : torch.nn.Module
        A Pytorch model.
    loss_fn : torch.nn.Module
        A Pytorch loss function.
    optimizer : torch.optim.Optimizer
        A Pytorch optimizer.
    device : torch.device
        A Pytorch device.
    max_iter : int
        The maximum number of iterations.
    \**kwargs : dict
        Additional keyword arguments.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        loss_fn: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        max_iter: int,
        **kwargs,
    ):
        self.kwargs = kwargs
        self.model = model
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.device = device
        self.max_iter = max_iter

    def fit(
        self,
        X: Union[DataLoader, torch.Tensor],
        y: Optional[Union[torch.Tensor, None]] = None,
    ):
        """
        fit() is a method that fits the model to the training data.

        Parameters
        ----------
        X : Union[DataLoader, torch.Tensor]
            The training data.
        y : Optional[Union[torch.Tensor, None]], default=None
            The training labels, only used if X is a ``torch.Tensor``.
        """

        # Check if X is a DataLoader
        if isinstance(X, DataLoader):
            if y is None:
                print("Training ...")
                self.model.train()

                # Iterate over epochs
                for epoch in range(self.max_iter):
                    print(f"Epoch {epoch+1}/{self.max_iter}")

                    # Iterate over batches
                    for _, (X_batch, y_batch) in enumerate(X):
                        self._fit(X_batch, y_batch)

            else:
                raise ValueError("If X is a DataLoader, y must be None")

        # Check if X is a torch Tensor
        elif isinstance(X, torch.Tensor):
            if y is None:
                raise ValueError("If X is a torch Tensor, y must not be None")

            # Check if y is a torch Tensor
            elif isinstance(y, torch.Tensor):
                print("Training ...")
                self.model.train()

                # Iterate over epochs
                for epoch in range(self.max_iter):
                    print(f"Epoch {epoch+1}/{self.max_iter}")
                    self._fit(X, y)

            else:
                raise ValueError("y must be a torch.Tensor")

        else:
            raise ValueError("X must be a DataLoader or a torch Tensor")

        return self

    def _fit(self, X, y):
        """
        _fit() is a private method that performs a forward pass, computes the loss
        and performs backpropagation.

        Parameters
        ----------
        X : torch.Tensor
            The training data.
        y : torch.Tensor
            The training labels.
        """
        X, y = X.to(self.device), y.to(self.device)

        # Forward pass
        pred = self.model(X)
        loss = self.loss_fn(pred, y)

        # Backpropagation
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

    def predict_proba(self, X: Union[DataLoader, torch.Tensor]):
        """
        predict_proba() is a method that predicts the probability of each class.

        Parameters
        ----------
        X : Union[DataLoader, torch.Tensor]
            The data to predict.
        """
        if X is None:
            raise ValueError("X must be a DataLoader or a torch Tensor")

        else:
            # check if X is a DataLoader
            if isinstance(X, DataLoader):
                print("Predicting ...")
                self.model.eval()
                predictions = []

                # Iterate over batches
                for _, (X_batch, _) in enumerate(X):
                    predictions_batch = self._predict_proba(X_batch)
                    predictions.append(predictions_batch)

                # Concatenate predictions
                predictions = torch.cat(predictions)
                return predictions

            # check if X is a torch Tensor
            elif isinstance(X, torch.Tensor):
                print("Predicting ...")
                self.model.eval()
                return self._predict_proba(X)

            else:
                raise ValueError("X must be a DataLoader or a torch Tensor")

    def _predict_proba(self, X):
        """
        _predict_proba() is a private method that predicts the probability
        of each class.

        Parameters
        ----------
        X : torch.Tensor
            The data to predict.
        """
        X = X.to(self.device)
        pred = self.model(X)
        probabilities = F.softmax(pred, dim=1)
        return probabilities

    def predict(self, X: Union[DataLoader, torch.Tensor]):
        """
        predict() is a method that predicts the class of each sample.

        Parameters
        ----------
        X : Union[DataLoader, torch.Tensor]
            The data to predict.
        """
        pred = self.predict_proba(X)
        return torch.argmax(pred, dim=1)



import numpy as np
from scipy.special import softmax


def get_exponential_probabilities(n, p=1.0, tau=1.0):
    """Get probabilities from exponential distribution for n classes.

    Parameters
    ----------
    n : int
            Number of classes.
    p : float, default=1.0
            Exponent parameter.
    tau : float, default=1.0
            Scaling parameter.

    Returns
    -------
    probs : 2d array-like
            Matrix of probabilities where each row represents the true class
            and each column the probability for class n.

    Example
    -------
    >>> get_exponential_probabilities(5)
    array([[0.63640865, 0.23412166, 0.08612854, 0.03168492, 0.01165623],
    [0.19151597, 0.52059439, 0.19151597, 0.07045479, 0.02591887],
    [0.06745081, 0.1833503 , 0.49839779, 0.1833503 , 0.06745081],
    [0.02591887, 0.07045479, 0.19151597, 0.52059439, 0.19151597],
    [0.01165623, 0.03168492, 0.08612854, 0.23412166, 0.63640865]])
    """

    probs = []

    for true_class in range(0, n):
        probs.append(-(np.abs(np.arange(0, n) - true_class) ** p) / tau)

    return softmax(np.array(probs), axis=1)



import numpy as np
from scipy.special import softmax
from scipy.stats import poisson


def get_poisson_probabilities(n):
    """Get probabilities from poisson distribution for n classes.

    Parameters
    ----------
    n : int
        Number of classes.

    Returns
    -------
    probs : 2d array-like
        Matrix of probabilities where each row represents the true class
        and each column the probability for class n.
    """

    probs = []

    for true_class in range(1, n + 1):
        probs.append(poisson.pmf(np.arange(0, n), true_class))

    return softmax(np.array(probs), axis=1)



import math

import numpy as np

from .utils import get_intervals, triangular_cdf


def get_triangular_probabilities(n: int, alpha2: float = 0.01, verbose: int = 0):
    """
    Get the probabilities for the triangular distribution.

    Parameters
    ----------
    n : int
        Number of classes.
    alpha2 : float, optional
        Alpha2 value, by default 0.01.
    verbose : int, optional
        Verbosity level, by default 0.
    """

    if verbose >= 1:
        print(f"Computing triangular probabilities for {n=} and {alpha2=}...")

    def compute_alpha1(alpha2):
        c_minus = (1 - 2 * alpha2) * (2 * alpha2 - math.sqrt(2 * alpha2))

        return pow((1 - math.sqrt(1 - 4 * c_minus)) / 2, 2)

    def compute_alpha3(alpha2):
        c1 = (
            pow((n - 1) / n, 2)
            * (1 - 2 * alpha2)
            * (math.sqrt(2 * alpha2) * (-1 + math.sqrt(2 * alpha2)))
        )

        return pow((1 - math.sqrt(1 - 4 * c1)) / 2, 2)

    alpha1 = compute_alpha1(alpha2)
    alpha3 = compute_alpha3(alpha2)

    if verbose >= 1:
        print(f"{alpha1=}, {alpha2=}, {alpha3=}")

    def b1(n):
        return 1.0 / ((1.0 - math.sqrt(alpha1)) * n)

    def m1(n):
        return math.sqrt(alpha1) / ((1 - math.sqrt(alpha1)) * n)

    def aj(n, j):
        num1 = 2.0 * j - 2 - 4 * j * alpha2 + 2 * alpha2
        num2 = math.sqrt(2 * alpha2)
        den = 2.0 * n * (1 - 2 * alpha2)

        max_value = (j - 1.0) / n

        # +-
        return (
            (num1 + num2) / den
            if (num1 + num2) / den < max_value
            else (num1 - num2) / den
        )

    def bj(n, j):
        num1 = 2.0 * j - 4 * j * alpha2 + 2 * alpha2
        num2 = math.sqrt(2 * alpha2)
        den = 2.0 * n * (1 - 2 * alpha2)

        min_value = j / n

        # +-
        return (
            (num1 + num2) / den
            if (num1 + num2) / den > min_value
            else (num1 - num2) / den
        )

    def aJ(n):
        aJ_plus = 1.0 + 1.0 / (n * (math.sqrt(alpha3) - 1.0))
        aJ_minus = 1.0 + 1.0 / (-n * (math.sqrt(alpha3) - 1.0))
        return aJ_plus if aJ_plus > 0.0 else aJ_minus

    def nJ(n):
        num = math.sqrt(alpha3)
        den = n * (1 - alpha3)
        return num / den

    if verbose >= 3:
        print(f"{b1(n)=}, {m1(n)=}, {aJ(n)=}, {nJ(n)=}, {aj(n, 1)=}, {bj(n,1)=}")
        for i in range(1, n + 1):
            print(f"{i=}  {aj(n, i)=}, {bj(n,i)=}")

    intervals = get_intervals(n)
    probs = []

    # Compute probability for each interval (class) using the distribution function.
    for j in range(1, n + 1):
        j_probs = []
        if j == 1:
            a = 0.0
            b = b1(n)
            c = 0.0
        elif j == n:
            a = aJ(n)
            b = 1.0
            c = 1.0
        else:
            a = aj(n, j)
            b = bj(n, j)
            c = (a + b) / 2.0

        if verbose >= 1:
            print(f"Class: {j}, {a=}, {b=}, {c=}, (j-1)/J={(j-1)/n}, (j/J)={j/n}")

        for interval in intervals:
            j_probs.append(
                triangular_cdf(interval[1], a, b, c)
                - triangular_cdf(interval[0], a, b, c)
            )
            if verbose >= 2:
                print(f"\tinterval: {interval}, prob={j_probs[-1]}")

        probs.append(j_probs)

    return np.array(probs)
